/*!
 * MouseIntent jQuery Plug-in
 *
 * Copyright 2013 Giva, Inc. (http://www.givainc.com/labs/) 
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * 	http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Date: 2013-10-03
 * Rev:  1.0.00
 */
;(function($){
	// create the plugin
	$.fn.mouseintent = function (options){
		var self = this;

		return this.each(function (i, el){
			var $el = self.eq(i)
				, mi = $el.data("mouseintent");
			// if the object isn't already initialized, then create a new instance
			if( !mi ){
				$el
					.data("mouseintent", new MouseIntent(self.eq(i), options))
					// when this event is called, make sure to remove the mouse event data on the element
					.bind("mouseintent.destroy", function (){
						$el.removeData("mouseintent");
					});
			}
		});
	};

	// initialize plugin info
	$.mouseintent = {
		// set the version of the plug-in
		  version: "1.0.00"
		// define default settings
		, defaults: {
			// define global settings
			  borderThreshold: 100
			, monitorMovement: true
			, monitorFrequency: 100
			, mouseAwayAfter: 1500
			
			// event handlers
			, init: $.noop                // occurs when the plugin is initialized
			, mouseaway: $.noop           // occurs when the user appears to purposefully moves mouse away from element
			, mousereenter: $.noop        // occurs when the mouse reenters the target area
		}
	};
	
	// global variables
	var Tracker = new Collection();
	
	// a generic collection to track items
	function Collection(items){
		this.items = items || [];
	}
	Collection.prototype.get = function(i){
		return this.items[i];
	}
	Collection.prototype.getAll = function(i){
		return this.items;
	}
	Collection.prototype.count = function(o){
		return this.items.length;
	}
	Collection.prototype.add = function(o){
		this.items.push(o);
	}
	Collection.prototype.remove = function(o){
		// find the object and remove
		for( var i=0; i < this.count(); i++ ){
			// we found the item to unregister
			if( o === this.items[i] ){
				this.items.splice(i, 1);
				return true;
			}
		}
		return false;
	}

	// global mouse tracking	
	var TrackMouse = {
		  positions: []
		, isTracking: false
		, start: function (){
			// if already tracking, stop
			if( TrackMouse.isTracking ) return;

			// start monitoring the mouse
			$(document).bind("mousemove.TrackMouse", TrackMouse.eventHandler);
			TrackMouse.isTracking = true;
		}
		, stop: function (){
			// stop monitoring the mouse
			$(document).unbind("mousemove.TrackMouse");
			TrackMouse.isTracking = false;
		}
		, get: function (){
			return {
				  current: TrackMouse.positions[TrackMouse.positions.length-1]
				, previous: TrackMouse.positions[0]
			}
		}
		, eventHandler: function (e){
			var event = {x: e.pageX, y: e.pageY, timeStamp: (new Date).getTime()};

			// store the position
			TrackMouse.positions.push(event);
			
			// only store last 2 positions
			if( TrackMouse.positions.length > 3 ){
				TrackMouse.positions.shift();
			}
		}
	};

	function MouseIntent(el, options){
		// a reference to the object
		var self = this, $el = el, dimensions;
		
		// make sure we're tracking mouse movement
		TrackMouse.start();
		
		// register object
		Tracker.add(self);
		
		// store the options
		this.options = options = $.extend({}, $.mouseintent.defaults, options);

		// normalize the border		
		normalizeBorderThreshold(this.options);
		
		// set public methods
		this.create = function(){
			// run initialization
			this.updateDimensions();

			$el.bind("mouseleave.mouseintent", function (){
				trackMouse();
			});
		}
		
		// set public methods
		this.destroy = function(){
			// unregister object
			Tracker.remove(self);
			// if all instances are destroyed
			if( Tracker.count() === 0 ){
				// stop tracking the mouse
				TrackMouse.stop();
			}
			// remove all events
			$el.unbind(".mouseintent");
			// make sure to remove any binding done by the plugin
			$el.trigger("mouseintent.destroy");
		}
		
		// run when the mouse goes away from the intended target
		this.mouseaway = function(){
			// indicate the plug-in has initialized
			options.mouseaway.apply(self, [options]);
			// fire an event
			$el.trigger("mouseintent-mouseaway", [options]);

			// kill the plugin
			this.destroy();
		}
		
		// run when the mouse goes away from the intended target
		this.mousereenter = function(){
			// indicate the plug-in has initialized
			options.mousereenter.apply(self, [options]);
			// fire an event
			$el.trigger("mouseintent-mousereenter", [options]);
		}
		
		// set public methods
		this.updateDimensions = function(){
			return dimensions = ElementHelper.getDimensions($el);
		}
		
		// set public methods
		function normalizeBorderThreshold(options){
			var bt = options.borderThreshold;
			// if value doesn't need normalizing
			if( $.isNumeric(bt) ){
				return;
			}

			// if we don't have an array, throw an error
			if( !$.isArray(bt) ){
				throw "An invalid borderThreshold has been specified. This value must either be a numeric value or an array.";
			}

			// 0 = top, 1 = right, 2 = bottom, 3 = left

			// make sure the array is the correct length/format
			if( bt.length === 1 ) bt = [bt[0], bt[0], bt[0], bt[0]];
			else if( bt.length === 2 ) bt = [bt[0], bt[1], bt[0], bt[1]];
			else if( bt.length === 3 ) bt = [bt[0], bt[1], bt[2], bt[1]];
			// return the specified border direction
			options.borderThreshold = bt;
		}
		
		function trackMouse(lp, started){
			var epoch = (new Date()).getTime()
				, mousePos = TrackMouse.get()
				, isFirstRun = !lp
				, fromEdge = getMouseFromEdge(mousePos.current, isFirstRun)
				/*
				 * we want to flag the mouse as moving if we're not tracking mouse movement *or* the
				 * mouse is being moved. we know we're moving if the last mouse movement time stamp 
				 * doesn't match and the current timestamp
				 */
				, isMoving = !self.options.monitorMovement || isFirstRun || (mousePos.current.timeStamp !== lp.timeStamp);

			// if we haven't flagged our starting time, start logging now			
			if( !started ) started = epoch;

			// make sure the mouse is still moving, we're inside our threshold area and we haven't mouse around too long			
			if( isMoving && fromEdge.insideThreshold && !hasReachedMaxDelay(started, epoch) ){
				// re-run the function to check again
				setTimeout(function (){
					// we run the function again, passing in the timestamp of the current mouse movement
					trackMouse(mousePos.current, started);
				}, self.options.monitorFrequency);
			// the mouse has re-entered the target area
			} else if( !isFirstRun && (fromEdge.distance === -1) ){
				self.mousereenter();
			// we've moved outside the target area
			} else {
				self.mouseaway();
			}
		}
		
		function getBorderThreshold(direction){
			// return the current threshold
			if( $.isNumeric(self.options.borderThreshold) ){
				return self.options.borderThreshold;
			} else {
				// 0 = top, 1 = right, 2 = bottom, 3 = left
				return self.options.borderThreshold[direction];
			}
		}
		
		function getMouseFromEdge(mousePos, firstRun){
			var distance = calculateDistanceFromEdge(dimensions, mousePos)
				, direction = calculateDirectionToCenter(dimensions, mousePos)
				, threshold = getBorderThreshold(direction.value);
				
			return {
				  distance: distance
				, direction: direction.value
				, threshold: threshold
				, isWithinThresholdContainer: isWithinThresholdContainer(dimensions, mousePos)
				/*
				 * - we have to have a threshold to be inside it
				 * - we always run on first run, because the mouse coordinates may not be current
				 * - we want to makes sure the distance is outside the target (> -1)
				 * - but it must also be within the threshold
				 * - and finally we want to make sure the mouse is within the rectangled target zone
				 */
				// we must always run on first run, because mouse coordinates may be out of date
				, insideThreshold: (threshold > 0) && ((firstRun === true) || ((distance > -1) && (distance <= threshold) && isWithinThresholdContainer(dimensions, mousePos)))
			};
		}

		function isWithinThresholdContainer(rect, mouse){
			var threshold = {
				  top: rect.top - getBorderThreshold(0)
				, right: rect.right + getBorderThreshold(1)
				, bottom: rect.bottom + getBorderThreshold(2)
				, left: rect.left - getBorderThreshold(3)
			};
			
			// make sure we're in the threshold rect
			return !(mouse.x < threshold.left || mouse.x > threshold.right || mouse.y < threshold.top || mouse.y > threshold.bottom);
		}

		function hasReachedMaxDelay(started, epoch){
			return (self.options.mouseAwayAfter > 0) && ((epoch-started) >= self.options.mouseAwayAfter);
		}
		
		// create the instance
		this.create();
		
		// indicate the plug-in has initialized
		options.init.apply(self, [options]);
		// fire an event
		$el.trigger("mouseintent-init", [options]);
		
		return this;
	};

	// gets the distances from the edge of an rectangle
	function calculateDistanceFromEdge(rect, mouse){
		var x=mouse.x, y=mouse.y;

		if( mouse.x < rect.left ){
			x = rect.left;
		} else if( mouse.x > rect.right ){
			x = rect.right;
		}

		if( mouse.y < rect.top ){
			y = rect.top;
		} else if( mouse.y > rect.bottom ){
			y = rect.bottom;
		}
		
		return ((x === mouse.x) && (y === mouse.y)) ? -1 : Math.floor(Math.sqrt(Math.pow(mouse.x - x, 2) + Math.pow(mouse.y - y, 2)));
	};

	/* http://stackoverflow.com/questions/3627042/jquery-animation-for-a-hover-with-mouse-direction/3647634#3647634 */
	function calculateDirectionToCenter(rect, mouse){   
		/*
		 * calculate the x and y to get an angle to the center of the div from that x and y.
		 * gets the x value relative to the center of the DIV and "normalize" it
		 */
		var x = (mouse.x - rect.left - (rect.width/2)) * (rect.width > rect.height ? (rect.height/rect.width) : 1)
			, y = (mouse.y - rect.top  - (rect.height/2)) * (rect.height > rect.width ? (rect.width/rect.height) : 1);
	
		/*
		 * the angle and the direction from where the mouse came in/went out clockwise (TRBL=0123)
		 * 
		 * - first calculate the angle of the point
		 * - add 180 deg to get rid of the negative values
		 * - divide by 90 to get the quadrant
		 * - add 3 and do a modulo by 4  to shift the quadrants to a proper clockwise TRBL (top/right/bottom/left)
		 */
		var direction = Math.round((((Math.atan2(y, x) * (180 / Math.PI)) + 180 ) / 90 ) + 3 )  % 4
			, label = (["top", "right", "bottom", "left"])[direction];
		
		return {value: direction, label: label};
	}


/*
	// gets the distances from the center of an rectangle
	function calculateDistanceFromCenter(rect, mouse){
		return Math.floor(Math.sqrt(Math.pow(mouse.x - (rect.left+(rect.width/2)), 2) + Math.pow(mouse.y - (rect.top+(rect.height/2)), 2)));
	}

	function distMetric(x, y, x2, y2){
		var xDiff = x - x2;
		var yDiff = y - y2;
		return (xDiff * xDiff) + (yDiff * yDiff);
	};
	
	function slope(a, b) {
		return (b.y - a.y) / (b.x - a.x);
	};
*/
	
	var ElementHelper = {
		getDimensions: function (el){
			var $el = $(el)
				, offset = $el.offset()
				, width = $el.outerWidth()
				, height = $el.outerHeight();

			return {
				  left: offset.left
				, right: offset.left + width
				, top: offset.top
				, bottom: offset.top + height
				, width: width
				, height: height
			};
		}
	};

})(jQuery);
